
import qualified Data.Set as S
import qualified Data.Map as M
import Data.Map ((!))
import Data.Maybe
import Data.List

type Cell = S.Set Int

type Coord = (Int,Int)
type Grid = M.Map Coord Cell

type Number = [Cell]
type Field = [Coord]

data Rule =	AB Field
		| Gt Coord Coord --c1 > c2
		| K Field
		| P Field
		| Eq Coord Coord --Cell equality
		| PZ Field
		| Q Field
		| U Field
		| NEq Coord Coord --Cell inequality
		| V Field Field
		| Gl Field Field
		| Brute

type Rules = [Rule]

data Problem = Problem {
	width  :: Int,
	height :: Int,
	pRules  :: Rules
	}

data State = State {
	grid :: Grid,
	sRules :: Rules
}

-- Grid functions

numbers = S.fromList [0..9]

initGrid w h = M.fromList $ do
	x <- [0..(h-1)]
	y <- [0..(w-1)]
	return ((x,y),numbers)

isSolved grid = M.null $ M.filter (not . isFixed) grid
isUnsolvable grid = not $ M.null $ M.filter S.null grid

showSolution grid = unlines $ map formatLine lines
  where	fields = M.toAscList grid
        lines  = groupBy (\a b -> fst (fst a) == fst (fst b)) fields
	formatLine line = concat $ map (show . head . S.elems . snd) line

readField grid field = foldl ((+).(*10)) 0  $ map (head . S.elems . (grid !)) field

-- Cell functions

isFixed s = S.size s == 1

-- Helper function
palinPairs [] = []
palinPairs l = take (length l `div` 2) $ zip l (reverse l)

firstJust = listToMaybe . catMaybes

prims = 2 : (filter (\n -> not (any (\p -> n `mod` p == 0) (takeWhile (< n`div`2) prims)))) [3..]

digits = reverse . digits'
digits' 0 = []
digits' n = n `mod` 10 : digits' (n `div` 10)

possibleDigits grid [] [] = True
possibleDigits grid [] _  = error "Trying wrong sized number"
possibleDigits grid _ []  = error "Trying wrong sized number"
possibleDigits grid (c:cs) (d:ds) | d `S.member` (grid ! c) = possibleDigits grid cs ds
                                  | otherwise               = False

setTo grid [] [] = grid
setTo grid [] _  = error "Trying wrong sized number"
setTo grid _ []  = error "Trying wrong sized number"
setTo grid (c:cs) (d:ds) = setTo (M.insert c (S.singleton d) grid) cs ds


splitSet s = S.partition (<= pivot) s
  where	max = S.findMax s
	min = S.findMin s
	pivot = (max + min) `div` 2

-- Test problems

problem1 = Problem 2 1 [P field, PZ field]
  where	field = [(0,0), (0,1)]

problem2 = Problem 2 1 [PZ field, AB field]
  where	field = [(0,0), (0,1)]

problem3 = Problem 3 1 [PZ field, P field]
  where	field = [(0,0), (0,1), (0,2)]

problem4 = Problem 3 1 [Q field, P field]
  where	field = [(0,0), (0,1), (0,2)]

problem5 = Problem 3 1 [Q field, AB field]
  where	field = [(0,0), (0,1), (0,2)]

problem6 = Problem 3 1 [Q field, K field]
  where	field = [(0,0), (0,1), (0,2)]

problem7 = Problem 3 1 [K field, AB field] -- no solution
  where	field = [(0,0), (0,1), (0,2)]

problem8 = Problem 3 1 [Q field, U field] 
  where	field = [(0,0), (0,1), (0,2)]

problem9 = Problem 3 1 [P field, U field]  -- no solution
  where	field = [(0,0), (0,1), (0,2)]

problem10 = Problem 2 1 [P field]
  where	field = [(0,0), (0,1)]

problem11 = Problem 3 2 [PZ field1, P field1, Gl field1 field2]
  where	field1 = [(0,0), (0,1), (0,2)]
  	field2 = [(1,0), (1,1), (1,2)]

problem12 = Problem 3 2 [PZ field1, P field1, V field2 field1, P field2]
  where	field1 = [(0,0), (0,1), (0,2)]
  	field2 = [(1,0), (1,1), (1,2)]

problem13 = Problem 3 2 [PZ field1, P field1, V field2 field1, P field2, PZ field3, V field4 field3]
  where	field1 = [(0,0), (0,1), (0,2)]
  	field2 = [(1,0), (1,1), (1,2)]
	field3 = [(0,0), (1,0)]
	field4 = [(1,0), (1,1)]

-- State function

startState (Problem w h rules) = State (initGrid w h) (rules++[Brute])

-- Solver

solve problem = case result of
		Just grid -> putStr $ showSolution grid
		Nothing   -> putStrLn $ "no solution"
  where result = solve' $ normalizeRules $ startState problem

normalizeRules (State grid rules) = State grid (concatMap normalizeRule rules)

normalizeRule (P l)  = map (uncurry Eq) (palinPairs l)
normalizeRule (AB l) = zipWith Gt l (tail l)
normalizeRule (U []) = []
normalizeRule (U (c:l)) = map (NEq c) l ++ normalizeRule (U l)
normalizeRule (Gl f1 f2) = zipWith Eq f1 f2 -- test for same size
normalizeRule r      = [r]

solve' s@(State grid rules)
	| isUnsolvable grid	= Nothing
	| null rules && isSolved grid = Just grid
	| null rules            = Nothing -- No more rules, but unsolved?
	| otherwise    		= applyRule s

applyRule (State grid (r:rs)) = firstJust (map (\g -> solve' (State g nextrules)) newgrids)
  where	(done, newgrids) = applyRule' grid r
        nextrules | done     = rs
	          | not done = rs ++ [r]

applyRule' grid (P _) = error "no normalizing"

applyRule' grid (Eq c1 c2) = (done, newgrids)
  where n1 = grid ! c1
        n2 = grid ! c2
	candits = S.intersection n1 n2
	done = isFixed candits
	newgrids = [M.insert c1 candits $ M.insert c2 candits $ grid]

applyRule' grid (Gt c1 c2) = (done, newgrids)
  where n1 = grid ! c1
        n2 = grid ! c2
	cand1 = S.filter (> S.findMin n2) n1
	cand2 = S.filter (< S.findMax n1) n2
	done = S.null cand1 || S.null cand2 || S.findMin cand1 > S.findMax cand2
	newgrids = [M.insert c1 cand1 $ M.insert c2 cand2 $ grid]

applyRule' grid (NEq c1 c2) = (done, newgrids)
  where n1 = grid ! c1
        n2 = grid ! c2
	done = S.null $ S.intersection n1 n2 
	newgrids | n1 == n2  = []
	         | otherwise = [grid]

applyRule' grid (PZ field) = pregen grid field prims
applyRule' grid (Q field) = pregen grid field $ map (^2) [1..]
applyRule' grid (K field) = pregen grid field $ map (^3) [1..]

-- quite stupid implementation, waits till second field is fixed
applyRule' grid (V f1 f2) = result
  where applicable = all isFixed $ map (grid !) f2
  	number = readField grid f2
	result | applicable = pregen grid f1 $ map (number*) [2..] -- 1.. wenn gleichheit erlaubt
	       | otherwise  = (False, [grid])


-- Catch all last rule, start brute-forcing
applyRule' grid Brute = (done, newgrids)
  where notdone set = S.size set > 1
  	notdones = M.toList $ M.filter notdone grid
	(c,s) = head $ notdones
	(s1,s2) = splitSet s
	done = isSolved grid
	newgrids | done           = [ grid ]
		 | null notdones  = []
	         | otherwise      = [ M.insert c s1 grid, M.insert c s2 grid ]


-- assumes an ascending list
pregen grid field list = (True, newGrids)
  where max = 10 ^ (length field) - 1
        min = 10 ^ (length field - 1)
	canditsN = takeWhile (<=max) $ dropWhile (<min) $ list
	canditsZ = map digits canditsN
	possibleCandits = filter (possibleDigits grid field) canditsZ
	newGrids = map (setTo grid field) possibleCandits

