# -*- coding: utf-8 -*- 

# Copyright (c) 2008 Manuel Holtgrewe, Mathias Ziebarth, Lucas Lürich, Joachim Breitner
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import struct
import mmap
import os

class BinList:
    """
    Helper Class. Loads a list of sorted binary stored long integers
    from a file and provides a fast (O(log n)) element check.
    """
    def __init__(self, filename):
        try:
            f = file(filename)
            size = os.path.getsize(filename)
            self.data = mmap.mmap(f.fileno(), size, access = mmap.ACCESS_READ)
        except:
            self.data = file(filename).read()
        self.format = "i"
        self.itemsize = struct.calcsize(self.format)
        assert len(self.data) % self.itemsize == 0

    def __len__(self):
        return len(self.data) / self.itemsize

    def __getitem__(self, i):
        if i<0:
            i = len(self)+i
        bin = self.data[ i * self.itemsize : (i+1) * self.itemsize ]
        (i,) = struct.unpack(self.format, bin)
        return i

    def __contains__(self, n):
        lo, hi = 0, len(self)-1
        while lo <= hi:
            mid = (lo + hi) / 2
            if self[mid] < n:
                lo = mid + 1
            elif n < self[mid]:
                hi = mid - 1
            else:
                return True
        return False

    class Iter:
        def __init__(self, l):
            self.l = l
            self.n = 0

        def __iter__(self): 
            return self

        def next(self):
            self.n += 1
            if self.n >= len(self.l):
                raise StopIteration
            else:
                return self.l[self.n]

    def __iter__(self):
        return BinList.Iter(self)



# vim:sw=4:ts=4:sts=4:expandtab
