# -*- coding: utf-8 -*- 

# Copyright (c) 2008 Manuel Holtgrewe, Mathias Ziebarth, Lucas Lürich, Joachim Breitner
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import string
import re

import sevendown.rules as sr
import sevendown.graph as sg

BORDER_BOTTOM = 'b'
BORDER_RIGHT  = 'r'

class SpecialCaseRuleFactory(object):
    """
    This class provides the logic that can combine two rules into a special
    case rule.
    
    """
    
    def rulesForNumber(self, number):
        """
        Returns a list of rule for the number's related rules where a special
        case rule is known.
        
        TODO: Later consider all global pairs of rules.
        
        """
        related_rules = number.getRelatedRules()
        new_rules = []
        
        for left_rule in related_rules:
            for right_rule in related_rules:
                if left_rule is right_rule: continue # skip dupes
                
                # TODO: factor out if, call to method
                # X has unique digits and cross sum of X is known
                if isinstance(left_rule, sr.UniqueDigitsRule) and \
                        isinstance(right_rule, sr.CrossSumRule):
                    rule = sr.CrossSumUniqueCombinationRule(cross_sum.arg1, 
                                cross_sum.arg2)
                    new_rules.append(rule)
        
        return new_rules
    

class ParserException(Exception):
    """Thrown when the file to be loaded is invalid"""
    pass

class MarkerOutOfFieldException(ParserException):
    """Thrown when marker not in Field"""


class Loader(object):
    """Use this class to load a graph from a file object.
    
    >>> loader = Loader()
    >>> loader.load(sys.stdin)
    
    """
    
    def load(self, f):
        """Loads the problem description from the given file like object.  
        This could for example be stdin.  After loading the description, the
        rules will be parsed and the Rule objects created.
        
        """
        # load the file
        try:
            self.height, self.width = self.loadDimensions(f)
            markers = self.loadMarkerCoordinates(f)
            horizontal_markers = self.loadMarkerNames(f)
            vertical_markers = self.loadMarkerNames(f)
            rule_strings = self.loadRuleLines(f)
            
            # parse all rules
            parser = PredicateParser()
            rules = (parser.parse(parser.tokenize(s)) for s in rule_strings)
            
            # Create the field with a FieldFactory.
            factory = FieldFactory()
            return factory.createField(self.height, self.width, markers,
                horizontal_markers, vertical_markers, rules)
        except AssertionError, e:
            raise
            raise ParserException(e)
    
    def safe_split(self, line, n):
        """
        Splits a line, throwing an exception if the number of elements is invalid
        """
        s = line.split()
        if len(s) != n:
            raise ParserException("Could not parse line: \"%s\", expected %d values." % (line, n))
        return s

    def safe_int(self, val):
        try:
            n = int(val)
        except ValueError:
            raise ParserException("Expected a single integer: %s " % val)
        return n

    def loadDimensions(self, f):
        """Loads height and width from the next line of the file like object,
        performs some basic checks and then returns (height, width).
        
        """
        line = f.readline()
        try:
            height, width = map(int, self.safe_split(line, 2))
        except ValueError:
            raise ParserException("First line not parsable as integers: \"%s\"" % line)
        
        if not (1 <= height <= 10 and
                1 <= width <= 10):
            raise ParserException("Dimensions %dx%d out of bouds (max. 10x10)" % (width, height))
        
        return height, width
    
    def loadMarkerCoordinates(self, f):
        """Loads the marker names from the file like objects, performs some
        basic checks and returns a dict with all the markers.
        
        The dict's key is the marker's name, the value the triple described
        below.
        
        A marker consists of a tuple with length 3.  The first entry contains 
        the marker name (uppercase character), the second and third entry the
        coordinate on the field (first Y then X coordinate).
        
        Throws exception if marker is positioned outside the field.
        
        """
        line = f.readline()
        n = self.safe_int(line)
        
        markers = {}
        
        for i in range(0, n):
            line = f.readline()
            # We use (x,y) order, but input is in (y,x) form
            (n, ys, xs) = self.safe_split(line, 3)
            y = int(ys) - 1
            x = int(xs) - 1
            
            assert n in string.ascii_uppercase
            assert 0 <= y < 11
            assert 0 <= x < 11
            
            if (y >= self.height) or (x >= self.width):
                raise MarkerOutOfFieldException(n + " out of Field")
            
            markers[n] = (n, x, y)
        
        return markers
    
    def loadMarkerNames(self, f):
        """Loads the marker names for the horizontal or vertical direction."""
        n = int(f.readline())
        
        result = tuple(f.readline().strip() for i in range(0, n))
        
        return result
    
    def loadRuleLines(self, f):
        """Loads all rules from the file like objects.  This will only load
        the lines as strings.  No parsing or Rule construction is done as
        of yet.
        
        """
        line = f.readline()
        n = self.safe_int(line)
        
        lines = tuple(f.readline().strip() for i in range(0, n))
        
        return lines
    

class PredicateParser(object):
    """
    This class allows to parse rules as given by the problem description.
    
    It creates tuple based parse trees but does not create any Rule objects.
    
    The supported grammer EBNF (whitespace does not matter):
    
    unary-op-name    = "qp" | "qs" | "r"
    unary-op         = unary-op-name, "(", operator, ",", operator, ")"
    
    binary-op-name   = "mal" | "minus" | "plus"
    binary-op        = binary-op-name, "(", operator, ",", operator, ")"
    
    field-identifier = "A_w" | ... | "Z_w" | "A_h" | ... | "Z_h"
    
    operator         = field-identifier | binary-op | unary-op
    
    unary-pred-name  = "AB" | "K" | "PAL" | "PZ" | "Q" | "U"
    unary-pred       = unary-pred-name, "(", operator, ")"
    
    binary-pred-name = "V" | "="
    binary-pred      = binary-pred-name, "(", operator, ",", operator ")"
    
    predicate        = binary-pred | unary-pred
    
    Note that the parser is not very efficient but should be fast enough to
    parse the small predicates from most inputs.
    
    """
    
    # Operator names.
    OP_NAMES = ("qp", "qs", "r", "mal", "minus", "plus")
    
    # Predicate names.
    PRED_NAMES = ("AB", "K", "PAL", "PZ", "Q", "U", "V", "=", ":=")
    
    # Symbol arity
    ARITY = {}
    for s in ("qp", "qs", "r", "AB", "K", "PAL", "PZ", "Q", "U"):
        ARITY[s]= 1
    for s in ("mal", "minus", "plus", "V", "=", ":="):
        ARITY[s]=2
    
    # Valid field identifiers.
    FIELD_IDENTIFIERS = tuple("%s_%s" % (c, direction) \
        for direction in ('w', 's') \
        for c in string.ascii_uppercase)
    
    def parse(self, tokens):
        """Parse the given sequence of tokens.
        
        The result will be a tree of tuples where each tuple represents either
        an unary/binary operator or unary/binary predicate.  The first entry
        contains the name of the operator or predicate.  The second and 
        possibly third entry contains the operator the predicate/operator is 
        applied on.
        
        You can determine the type of a node, e.g. by "ARITY[node[0]]"
        Leafs are one-element long tuples.
        
        Example:
        
        >>> parser = PredicateParser()
        >>> tokens = ("=", "(", "A_s", "mal", "(", "B_s", "H_w", ")", ")")
        >>> parser.parse(tokens)
        ('=', ('A_s',), ('mal', ('B_s',), ('H_w',)))
        
        """
        result, ignored = self.parsePredicate(tokens)
        return result
    
    def parsePredicate(self, tokens):
        """Parses and returns a predicate as well as the yet unparsed elements
        from tokens.
        
        """
        assert len(tokens) >= 4, "Not enough tokens: %s"%str(tokens)
        assert tokens[0] in self.PRED_NAMES, "%s not a Predicate" % tokens[0]
        assert tokens[1] == '('
        assert tokens[-1] == ')'
        
        if PredicateParser.ARITY[tokens[0]] == 1:
            arg1, rest_tokens = self.parseOperator(tokens[2:])
            
            assert len(rest_tokens) > 0
            assert rest_tokens[0] == ')'
            
            return (tokens[0], arg1), rest_tokens[1:]
        elif PredicateParser.ARITY[tokens[0]] == 2:
            arg1, rest_tokens = self.parseOperator(tokens[2:])
            arg2, rest_tokens = self.parseOperator(rest_tokens)
            
            assert len(rest_tokens) > 0
            assert rest_tokens[0] == ')'
            
            return (tokens[0], arg1, arg2), rest_tokens[1:]
    
    def parseOperator(self, tokens):
        """Parses and and returns an operator."""
        assert (tokens[0] in self.OP_NAMES + self.FIELD_IDENTIFIERS) or \
            (re.compile('^[0-9]+$').match(tokens[0])), "Could not parse %s"%tokens[0]
        
        if tokens[0] in self.OP_NAMES:
            assert len(tokens) >= 4
            assert tokens[1] == '('
            assert tokens[-1] == ')'
            
            if self.ARITY[tokens[0]] == 1:
                arg1, rest_tokens = self.parseOperator(tokens[2:])
                
                assert len(rest_tokens) > 0
                assert rest_tokens[0] == ')'
                
                return (tokens[0], arg1), rest_tokens[1:]
            else: # tokens[0] in self.BINARY_OP_NAMES
                arg1, rest_tokens = self.parseOperator(tokens[2:])
                arg2, rest_tokens = self.parseOperator(rest_tokens)
                
                assert len(rest_tokens) > 0
                assert rest_tokens[0] == ')', rest_tokens
                
                return (tokens[0], arg1, arg2), rest_tokens[1:]
        elif tokens[0] in self.FIELD_IDENTIFIERS:
            # tokens[0] is field identifier
            return ((tokens[0],), tokens[1:])
        else: # tokens[0] must be number
            assert re.compile('^[0-9]+$').match(tokens[0])
            return ((int(tokens[0]),), tokens[1:])
    
    def tokenize(self, s):
        """Tokenize the given string."""
        # we do this by a very simple regular expression
        re_str = '(?:[a-z]+)|(?:[A-Z]_[ws])|(?:[A-Z]+)|=|\(|\)|(?:\d+)'
        p = re.compile(re_str)
        return tuple(p.findall(s))
    

class FieldFactory(object):
    """
    This class is a factory that creates a new Field object by the
    information which is loaded from a problem description.
    
    """
    
    def __init__(self):
        self.newCellNames = []
        """Stores the new temporary numbers."""
    
    def createField(self, height, width, markers, horizontal_markers,
            vertical_markers, rule_parsetrees):
        """Create the field.
        
        rules
          Parse tree represented as nested tuples; as returned by
          PredicateParser::parse().
        
        """
        # create the field
        field = sg.Field(height, width)
        
        field.markers = markers
        
        field.schema = self.calculateSchema(field, markers,
            horizontal_markers, vertical_markers)
        
        # add all horizontal and vertical numbers
        self.addNumbers(field, markers, horizontal_markers, True)
        self.addNumbers(field, markers, vertical_markers, False)
        
        flattened = []
        for rule_parsetree in rule_parsetrees:
            # flatten parse tree
            flattened += self.flattenParseTree(rule_parsetree)
        
        # add intermediate numbers
        self.addIntermediateNumbers(field)
        
        # create rules from parse tree
        for rule_simple_parsetree in flattened:
            rule = self.createRuleFromParseTree(rule_simple_parsetree, field)
            # normalize rules
            normalized = rule.normalize()
            for norm_rule in normalized:
                field.registerRule(norm_rule)
        
        # add rules per number
        for number in field.getNumbers():
            field.registerRule(sr.NoLeadingZeroRule(number))
            field.registerRule(sr.NumberRule(number))
        
        # for each number in the field
        #   do get the related rules R
        #      if there is a subset of R for which an efficient additional 
        #          rule is known then create this efficient rule.
        factory = SpecialCaseRuleFactory()
        for number in field.getNumbers():
            # SpecialCaseRuleFactory encapsulates the "if there is a
            # subset..." functionality
            new_rules = factory.rulesForNumber(number)
            for rule in new_rules:
                field.registerRule(rule)
        
        return field
    
    def calculateSchema(self, field, markers, horizontal_markers, 
            vertical_markers):
        """
        The field schema is the description of the stop markers on the
        field.  The field schema is a two dimensional array.  The first
        index selects the row in the field, the second the column.  The
        cells contain lists which contain all BORDER_* values that apply
        for the given field.
        
        """
        schema = tuple(tuple([] for j in range(field.width)) for i in \
            range(field.height))
        
        for cell in schema[-1]:
            cell.append(BORDER_BOTTOM)
        for col in schema:
            col[-1].append(BORDER_RIGHT)
        
        for col in range(field.height-1):
            for row in range(field.width):
                for (n,x,y) in markers.values():
                    if y == col+1 and x == row and n in vertical_markers:
                        schema[col][row].append(BORDER_BOTTOM)
        
        for col in range(field.height):
            for row in range(field.width-1):
                for (n,x,y) in markers.values():
                    if y == col and x == row +1 and n in horizontal_markers:
                        schema[col][row].append(BORDER_RIGHT)
        
        return schema
    
    def addNumbers(self, field, markers, marker_names, horizontal):
        """
        Adds all numbers for the given marker names to the fields in the
        both directions.
        
        """
        for key in marker_names:
            assert markers.has_key(key)
            # construct name
            if horizontal:
                dir = "w"
            else:
                dir = "s"
            name = "%s_%s" % (key, dir)
            
            pos_x, pos_y = markers[key][1:]
            length = self.getNumberLength(pos_x, pos_y, horizontal, field)
            
            number = field.createNumber(name, pos_x, pos_y, length,
                horizontal)
    
    def addIntermediateNumbers(self, field):
        """Adds all an IntermediateNumber for each name in
        self.newCellNames to the field.
        
        """
        for name in self.newCellNames:
            field.createIntermediateNumber(name)
    
    def getNumberLength(self, pos_x, pos_y, horizontal, field):
        """Returns the length of the number starting at the given position
        on the given field schema.  Number is horizontal iff horizontal is
        True.
        
        """
        length = 0
        
        if horizontal:
            while not BORDER_RIGHT in field.schema[pos_y][pos_x + length]:
                length += 1
            return length + 1
        else:
            while not BORDER_BOTTOM in field.schema[pos_y + length][pos_x]:
                length += 1
            return length + 1
    
    def newCellName(self):
        """Creates the name for a new temporary variable.  The names are made
        unique for the current loader by appending numbers to the prefix "_X".
        
        """
        name = "X_%d" % len(self.newCellNames)
        self.newCellNames.append(name)
        return (name,)
    
    def flattenParseTree(self, parseTree):
        newTrees = []
        ret = self.flattenParseTreeRec(parseTree, newTrees, True)
        if ret:
            newTrees.append(ret)
        return newTrees
    
    def flattenParseTreeRec(self, parseTree, newTrees, expectPredicate):
        """Given a parse tree (nested tuple representation), returns the
        new, flattened parse tree and a list of the new cells.
        
        New Cells will be called X_0, X_1, ...
        """
        op = parseTree[0]
        params = parseTree[1:]
        
        # Check op arity and type
        if expectPredicate:
            assert op in PredicateParser.PRED_NAMES, \
                "Excepted predicate, but %s is no predicate" % op
        else:
            assert op in PredicateParser.OP_NAMES, \
                "Excepted operator, but %s is no operator" % op
        
        assert PredicateParser.ARITY[op] == len(params), \
                "Symbol %s appearing with %d parameters, expected %d" % (op,
                    len(params), PredicateParser.ARITY[op])
        
        if op == "=":
            cell_name = None
            # We want to short-cut intermediate numbers
            for param in params:
                if len(param) == 1:
                    if cell_name:
                        # Both arguments are simple, just do nothing
                        return parseTree
                    else:
                        cell_name = param
            
            # Only create an interemdiate if all arguments are proper terms:
            if not cell_name:
                cell_name = self.newCellName()
            
            newparams = [op]
            for param in params:
                if len(param) > 1:
                    flatParam = self.flattenParseTreeRec(param, newTrees,
                        False)
                    newTrees.append((':=', cell_name, flatParam))
            return None
        else:
            newparams = [op]
            for param in params:
                if len(param) > 1:
                    cell_name = self.newCellName()
                    flatParam = self.flattenParseTreeRec(param, newTrees,
                        False)
                    newTrees.append((':=', cell_name, flatParam))
                    
                    newparams.append(cell_name)
                else:
                    newparams.append(param)
            return tuple(newparams)
    
    def createRuleFromParseTree(self, rule_parsetree, field):
        """Creates a new rule from the given rule parse tree.
        
        The parse tree must be a simple parse tree as returned by
        flattenParseTree().  All cells used in the rule must be already
        available in field.
        
        Returns the Rule instance for the given parse tree.
        """
        assert len(rule_parsetree) >= 2, "rule_parsetree is too short: %s" % \
            rule_parsetree
        assert rule_parsetree[0] in PredicateParser.PRED_NAMES, \
            "op \"%s\" is no predicate" % rule_parsetree[0]
        
        pred_name = rule_parsetree[0]
        
        if pred_name == ":=":
            assert len(rule_parsetree) == 3
            target = rule_parsetree[1]
            op = rule_parsetree[2][0]
            assert op in PredicateParser.OP_NAMES, \
                "op \"%s\" is no operator" % op
            args = rule_parsetree[2][1:]
            assert PredicateParser.ARITY[op] == len(args)
            
            numbers = []
            numbers.append(self.numberOrFixed(target[0], field))
            for arg in args:
                numbers.append(self.numberOrFixed(arg[0], field))
            rule = sr.createOperatorRule(op, *numbers)
        else:
            assert PredicateParser.ARITY[pred_name] == len(rule_parsetree) - 1
            
            numbers = []
            for arg in rule_parsetree[1:]:
                assert len(arg) == 1
                numbers.append(self.numberOrFixed(arg[0], field))
            rule = sr.createRule(pred_name, *numbers)
        
        return rule
    
    def numberOrFixed(self, name_or_int, field):
        """
        If name_or_int is an integer then it returns a new Number object
        which all digits set according to the for the given integer.
        
        Otherwise, it returns the Number object from field with this name.
        
        """
        if isinstance(name_or_int, int):
            fixed = sg.Fixed(name_or_int)
            fixed.field = field
            field.cells.append(fixed)
            return fixed
        else:
            return field.numberForName(name_or_int)


# vim:sw=4:ts=4:sts=4:expandtab
