#! /usr/bin/env python
# -*- coding: utf-8 -*- 

# Copyright (c) 2008 Manuel Holtgrewe, Mathias Ziebarth, Lucas Lürich, Joachim Breitner
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""
This module allows to store attributes of objects and reset them later.

The attributes are arranged in (linear) revisions.  The behaviour is similar
to a revision control system.

Example:

class Foo():
    def __init__(self):
        bar = 'bar!'
    
    @undoable('bar')
    def setBar(self, newBar):
        self.bar = newBar

foo = Foo()
foo.bar                 # => 'bar!'
foo.setBar('rex')
foo.bar                 # => 'rex'

revert()

foo.bar                 # => 'bar!'

"""

import copy

STATE_INITIAL = 'initial'
"""Module is reset/not set up yet."""
STATE_ACTIVE = 'active'
"""Module is started."""
STATE = STATE_INITIAL
"""Current module state."""

STORAGE = None
"""The revision storage.  Last entry is the current revision."""

def setUp():
    """
    Initialize the module's state.
    
    Creates the initial revision.  Must be called before startRevision(), 
    revert(), registerAttribute() or even tearDown() can be called.
    
    """
    global STATE, STORAGE, STATE_INITIAL, STATE_ACTIVE
    assert STATE == STATE_INITIAL
    STORAGE = [{}]
    STATE = STATE_ACTIVE


def tearDown():
    """
    Call this to reset the undo module.
    
    """
    global STATE, STORAGE, STATE_INITIAL, STATE_ACTIVE
    assert STATE == STATE_ACTIVE, "Undo manager in state %s, expected %s" % (str(STATE), str(STATE_ACTIVE))
    STORAGE = None
    STATE = STATE_INITIAL


def startRevision():
    """
    Starts a new revision and returns the number of current revisions.
    
    """
    global STATE, STORAGE, STATE_INITIAL, STATE_ACTIVE
    assert STATE == STATE_ACTIVE
    STORAGE.append({})
    return len(STORAGE)


def registerAttribute(obj, attr_name):
    """
    Register the given attribute with the given name and value for the object.
    
    getattr(obj, attr_name) can only be an immutable value or a builtin
    container of one.
    
    """
    global STATE, STORAGE, STATE_INITIAL, STATE_ACTIVE
    # Is should be possible to work with this method even when we are not
    # under undo's control
    if STATE != STATE_ACTIVE:
        return
    if not STORAGE[-1].has_key(obj):
        STORAGE[-1][obj] = {}
    if not STORAGE[-1][obj].has_key(attr_name):
        attr = getattr(obj, attr_name)
        
        if attr == None:
            pass # OK
        elif isinstance(attr, (float, int, str, bool)):
            pass # OK
        elif isinstance(attr, (dict, set, list)):
            # needs SHALLOW copying
            attr = copy.copy(attr)
        else:
            assert False, "%s has invalid type %s" % (attr, type(attr))
        
        STORAGE[-1][obj][attr_name] = attr


def revisionNo():
    """
    Returns the number of the current revision.
    
    """
    global STATE, STORAGE, STATE_INITIAL, STATE_ACTIVE
    assert STATE == STATE_ACTIVE
    assert len(STORAGE) > 0
    
    return len(STORAGE)


def revert():
    """
    Go back one revision.
    
    Returns number of revisions afterwards.
    
    """
    global STATE, STORAGE, STATE_INITIAL, STATE_ACTIVE
    assert STATE == STATE_ACTIVE
    assert len(STORAGE) > 0
    
    rev = STORAGE.pop()
    for obj, attrs in rev.iteritems():
        for attr_name, attr_value in attrs.iteritems():
            setattr(obj, attr_name, attr_value)
    
    return len(STORAGE)


def undoable(*attr_names):
    """
    Python decorator function that marks a method as "undoable".
    
    You have to specify the names of attributes.  These attributes of the
    object the method is called on will be stored in the current revision
    (if not present already) whenever the method is called.
    
    """
    global STATE, STORAGE, STATE_INITIAL, STATE_ACTIVE
    def decorate(func):
        """
        Returns a function that stores the attributes in attr_names before
        calling func.
        
        """
        def new_f(self, *args, **kwargs):
            """Call func with this arguments but store attributes before."""
            
            for attr_name in attr_names:
                registerAttribute(self, attr_name)
            
            return func(self, *args, **kwargs)
        return new_f
    
    return decorate

