#! /usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2008 Manuel Holtgrewe, Mathias Ziebarth, Lucas Lürich, Joachim Breitner
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import unittest

import sevendown.operators as so
import sevendown.graph as sg

class DummyPossibleValuesNumber(object):
    """A dummy number that returns predictable possible values."""
    
    def __init__(self, values):
        """Sets self.possible_values to values."""
        self.possible_values = values
    
    def possibleValues(self):
        """Returns self.possible_values."""
        return self.possible_values

    def isDone(self):
        return len(self.possible_values) == 1
    

class ModuleTestCase(unittest.TestCase):
    """Tests very basic properties of the rules module."""

class BinaryOperatorTestCase(unittest.TestCase):
    """Simple tests for binary operators."""
    
    def testInstantiationSetsCorrectAttributes(self):
        op = so.BinaryOperator(1, 2)
        
        self.assertEqual(op.argument1, 1)
        self.assertEqual(op.argument2, 2)
        self.assertEqual(op.name, 'UNKNOWN')
    
    def testStringConversionWorksCorrectly(self):
        op = so.BinaryOperator(1, 2)
        
        self.assertEqual(str(op), 'UNKNOWN(1, 2)')
    
class UnaryOperatorTestCase(unittest.TestCase):
    """Simple tests for binary operators."""
    
    def testInstantiationSetsCorrectAttributes(self):
        op = so.UnaryOperator(1)
        
        self.assertEqual(op.argument, 1)
        self.assertEqual(op.name, 'UNKNOWN')
    
    def testStringConversionWorksCorrectly(self):
        op = so.UnaryOperator(1)
        
        self.assertEqual(str(op), 'UNKNOWN(1)')
    
    def testIsNotDone(self):
        a = DummyPossibleValuesNumber([10, 11, 12, 13, 14])
        op = so.UnaryOperator(a)
        
        self.failIf(op.isDone())

    def testIsDone(self):
        a = DummyPossibleValuesNumber([1])
        op = so.UnaryOperator(a)
        
        self.assert_(op.isDone())


class CrossProductOperatorTest(unittest.TestCase):
    """Simple tests for the the CrossProductOperator."""
    
    def testInstantiationSetsCorrectAttributes(self):
        number = sg.Number('A_w', 1)
        op = so.CrossProductOperator(number)
        
        self.assertEqual(str(op), 'qp(A_w)')
        
    def testExecute(self):
        number = None
        op = so.CrossProductOperator(number)
        self.assertEqual(op.execute(120), 0)        
    
    def testValue(self):
        number = sg.Fixed(12345)
        op = so.CrossProductOperator(number)
        self.assertEqual(op.value(), 120)
    

class CrossSumOperatorTest(unittest.TestCase):
    """Simple tests for the the CrossSumOperator."""
    
    def testInstantiationSetsCorrectAttributes(self):
        number = sg.Number('A_w', 1)
        op = so.CrossSumOperator(number)
        
        self.assertEqual(str(op), 'qs(A_w)')
    
    def testExecute(self):
        number = None
        op = so.CrossSumOperator(number)
        self.assertEqual(op.execute(120), 3)
    
    def testValue(self):
        number = sg.Fixed(12345)
        op = so.CrossSumOperator(number)
        self.assertEqual(op.value(), 15)
    
    def testPossibleValues(self):
        a = DummyPossibleValuesNumber([10, 11, 12, 13, 14])
        op = so.CrossSumOperator(a)
        
        # list() converts a generator (or tuple, or another list) to a list
        # (list() is the constructor of the list type)
        result = list(map(op.execute, op.possibleValues()))
        expected = range(1, 6)
        
        self.assertEqual(result, expected)
    

class FlipOperatorTest(unittest.TestCase):
    """Simple tests for the the FlipOperator."""
    
    def testInstantiationSetsCorrectAttributes(self):
        number = sg.Number('A_w', 1)
        op = so.FlipOperator(number)
        
        self.assertEqual(str(op), 'r(A_w)')

    def testExecute(self):
        number = None
        op = so.FlipOperator(number)
        self.assertEqual(op.execute(120), 21)        
    
    def testValue(self):
        number = sg.Fixed(12345)
        op = so.FlipOperator(number)
        self.assertEqual(op.value(), 54321)


class MultiplicationOperatorTest(unittest.TestCase):
    """Simple tests for the the MultiplicationOperator."""
    
    def testInstantiationSetsCorrectAttributes(self):
        number = sg.Number('A_w', 1)
        number2 = sg.Number('B_w', 1)
        op = so.MultiplicationOperator(number, number2)
        
        self.assertEqual(str(op), 'mal(A_w, B_w)')
    
    def testExecute(self):
        number1 = None
        number2 = None
        op = so.MultiplicationOperator(number1, number2)
        self.assertEqual(op.execute(3,4), 12)        
    
    def testValue(self):
        number1 = sg.Fixed(100)
        number2 = sg.Fixed(5)
        op = so.MultiplicationOperator(number1, number2)
        self.assertEqual(op.value(), 500)

    def testPossibleValues(self):
        a1 = DummyPossibleValuesNumber([1, 10])
        a2 = DummyPossibleValuesNumber([2, 100])
        op = so.MultiplicationOperator(a1, a2)
        
        result = list(op.possibleValues())
        expected = [2,100,20,1000]
        
        self.assertEqual(result, expected)

class AdditionOperatorTest(unittest.TestCase):
    """Simple tests for the the AdditionOperator."""
    
    def testInstantiationSetsCorrectAttributes(self):
        number = sg.Number('A_w', 1)
        number2 = sg.Number('B_w', 1)
        op = so.AdditionOperator(number, number2)
        
        self.assertEqual(str(op), 'plus(A_w, B_w)')
    
    def testExecute(self):
        number1 = None
        number2 = None
        op = so.AdditionOperator(number1, number2)
        self.assertEqual(op.execute(3,4), 7)        
    
    def testValue(self):
        number1 = sg.Fixed(100)
        number2 = sg.Fixed(5)
        op = so.AdditionOperator(number1, number2)
        self.assertEqual(op.value(), 105)


class SubstractionOperatorTest(unittest.TestCase):
    """Simple tests for the the SubstractionOperator."""
    
    def testInstantiationSetsCorrectAttributes(self):
        number = sg.Number('A_w', 1)
        number2 = sg.Number('B_w', 1)
        op = so.SubstractionOperator(number, number2)
        
        self.assertEqual(str(op), 'minus(A_w, B_w)')
        
    def testExecute(self):
        number1 = None
        number2 = None
        op = so.SubstractionOperator(number1, number2)
        self.assertEqual(op.execute(4,3), 1)        
    
    def testValue(self):
        number1 = sg.Fixed(100)
        number2 = sg.Fixed(5)
        op = so.SubstractionOperator(number1, number2)
        self.assertEqual(op.value(), 95)
    

#
# Execute unit tests when called standalone.
if __name__ == "__main__":
    unittest.main()
 
# vim:sw=4:ts=4:sts=4:expandtab
