#! /usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2008 Manuel Holtgrewe, Mathias Ziebarth, Lucas Lürich, Joachim Breitner
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import unittest

import test_utils

import sevendown.undo as undo

class Foo(object):
    """Dummy class."""
    
    def __init__(self, value):
        self.bar = value
    
    @undo.undoable('bar')
    def setBar(self, bar):
        self.bar = bar
    

class NoLeadingZeroRuleTestCase(unittest.TestCase):
    def setUp(self):
        undo.setUp()
    
    def tearDown(self):
        undo.tearDown()
    
    def testFirstRegisterCounts(self):
        class Foo(object):
            pass
        
        foo = Foo()
        foo.bar = 'bar'
        undo.registerAttribute(foo, 'bar')
        self.assertEqual(foo.bar, 'bar')
        
        foo.bar = 'baz'
        undo.registerAttribute(foo, 'bar')
        self.assertEqual(foo.bar, 'baz')
        
        foo.bar = 'gnark'
        undo.registerAttribute(foo, 'bar')
        self.assertEqual(foo.bar, 'gnark')
        
        undo.revert()
        self.assertEqual(foo.bar, 'bar')
        
    
    def testRevisionNo(self):
        """Test the revisionNo() method."""
        self.assertEqual(undo.revisionNo(), 1)
        undo.startRevision()
        self.assertEqual(undo.revisionNo(), 2)
        undo.startRevision()
        self.assertEqual(undo.revisionNo(), 3)
        undo.revert()
        self.assertEqual(undo.revisionNo(), 2)
        undo.revert()
        self.assertEqual(undo.revisionNo(), 1)
    
    def testRevertWorks(self):
        foo = Foo(123)
        self.assertEqual(123, foo.bar)
        
        foo.setBar(42)
        self.assertEqual(42, foo.bar)
        
        self.assertEqual(undo.revert(), 0)
        self.assertEqual(123, foo.bar)
    
    def testNestedRevisions(self):
        foo = Foo(123)
        self.assertEqual(123, foo.bar)
        
        foo.setBar(42)
        self.assertEqual(42, foo.bar)
        
        undo.startRevision()
        foo.setBar(43)
        self.assertEqual(43, foo.bar)
        
        self.assertEqual(undo.revert(), 1)
        self.assertEqual(42, foo.bar)
        
        self.assertEqual(undo.revert(), 0)
        self.assertEqual(123, foo.bar)
    

# Execute unit tests when called standalone.
if __name__ == "__main__":
    unittest.main()

# vim:sw=4:ts=4:sts=4:expandtab
